import copy
import loguru
from openamundsen import (
    conf,
    constants,
    errors,
    fileio,
    meteo,
    modules,
    surface,
    statevars,
    terrain,
    util,
)
from .landcover import LandCover
import numpy as np
import pandas as pd
import rasterio
import sys
import time


class OpenAmundsen:
    """
    Class encapsulating the required data and methods for a single
    openAMUNDSEN model run.

    Parameters
    ----------
    config : dict
        Model run configuration.

    Examples
    --------
    >>> model = oa.OpenAmundsen(oa.read_config('config.yml'))
    >>> model.initialize()
    >>> model.run()
    """

    def __init__(self, config):
        self.config = conf.parse_config(config)
        self.logger = None
        self.state = None
        self.dates = None
        self.date = None
        self.date_idx = None

    def initialize(self):
        """
        Initialize the model according to the given configuration, i.e. read
        the required input raster files and meteorological input data,
        initialize the model grid and all required state variables, etc.
        """
        config = self.config

        self.require_soil = config.snow.model == 'multilayer'
        self.require_energy_balance = config.snow.melt.method == 'energy_balance'
        self.require_temperature_index = not self.require_energy_balance
        self.require_canopy = config.canopy.enabled
        self.require_evapotranspiration = config.evapotranspiration.enabled
        self.require_land_cover = self.require_canopy or self.require_evapotranspiration
        self.require_soil_texture = self.require_evapotranspiration

        self._initialize_logger()

        self._prepare_time_steps()
        self._initialize_grid()
        self._initialize_state_variable_management()

        if config.snow.model == 'multilayer':
            self.snow = modules.snow.MultilayerSnowModel(self)
        elif config.snow.model == 'cryolayers':
            self.snow = modules.snow.CryoLayerSnowModel(self)
        else:
            raise NotImplementedError

        if self.require_land_cover:
            self.land_cover = LandCover(self)

        if self.require_canopy:
            self.canopy = modules.canopy.CanopyModel(self)

        if self.require_evapotranspiration:
            self.evapotranspiration = modules.evapotranspiration.EvapotranspirationModel(self)

        # Create snow redistribution factor state variables
        for precip_corr in config.meteo.precipitation_correction:
            if precip_corr['method'] == 'srf':
                self.state.base.add_variable('srf', '1', 'Snow redistribution factor')
                break  # multiple SRFs are not allowed

        if config.snow_management.enabled:
            try:
                import openamundsen_snowmanagement
            except ImportError:
                raise errors.ConfigurationError('The snow management module must be installed '
                                                'for enabling snow management.')
            self.snow_management = openamundsen_snowmanagement.SnowManagementModel(self)

        self._create_state_variables()

        self._read_input_data()
        self._read_meteo_data()
        self._calculate_terrain_parameters()

        config.results_dir.mkdir(parents=True, exist_ok=True)  # create results directory if necessary
        self._initialize_point_outputs()
        self._initialize_gridded_outputs()

        self._initialize_state_variables()

        if self.config.liveview.enabled:
            self.logger.info('Creating live view window')
            from openamundsen import liveview
            lv = liveview.LiveView(self.config.liveview, self.state, self.grid.roi)
            lv.create_window()
            self.liveview = lv

    def run(self):
        """
        Start the model run. Before calling this method, the model must be
        properly initialized by calling `initialize()`.
        """
        self.logger.info('Starting model run')
        start_time = time.time()

        for _ in range(len(self.dates)):
            self.run_single()

        time_diff = pd.Timedelta(seconds=(time.time() - start_time))
        self.logger.success('Model run finished. Runtime: ' + str(time_diff))

    def run_single(self):
        """
        Process the next time step, i.e., increment the date counter and call the methods for
        preparing the meteorological fields and the interface to the submodules.
        """
        if self.date_idx is None:
            self.date_idx = 0
        elif self.date_idx == len(self.dates) - 1:
            raise errors.RuntimeError('Model run already finished')
        else:
            self.date_idx += 1

        self.date = self.dates[self.date_idx]

        self.logger.info(f'Processing time step {self.date:%Y-%m-%d %H:%M}')

        meteo.interpolate_station_data(self)
        self._process_meteo_data()
        self._model_interface()
        self.point_output.update()
        self.gridded_output.update()

        if self.config.liveview.enabled:
            self.logger.debug('Updating live view window')
            self.liveview.update(self.date)

    def global_mask(self, mask, global_mask=None, global_idxs=None):
        if global_idxs is None:
            if global_mask is None:
                global_idxs = self.grid.roi_idxs_flat
            else:
                global_idxs = np.flatnonzero(global_mask)

        if mask.shape[-1] != len(global_idxs):
            raise Exception('Local mask does not match global mask size')

        if mask.ndim == 1:
            global_mask = np.zeros((self.grid.rows, self.grid.cols), dtype=bool)
            idxs = global_idxs[mask]
            global_mask.flat[idxs] = True
        elif mask.ndim == 2:
            dim3 = mask.shape[0]
            global_mask = np.zeros((dim3, self.grid.rows, self.grid.cols), dtype=bool)

            for i in range(dim3):
                idxs = global_idxs[mask[i, :]]
                global_mask[i, :, :].flat[idxs] = True

        return global_mask

    def roi_mask_to_global(self, mask):
        return self.global_mask(mask)

    def _prepare_time_steps(self):
        """
        Prepare the time steps for the model run according to the run
        configuration (start date, end date, and time step).
        """
        dates = pd.date_range(
            start=self.config['start_date'],
            end=self.config['end_date'],
            freq=self.config['timestep'],
        )
        self.dates = dates

        # Store timestep in seconds in the `timestep` attribute
        self.timestep = util.offset_to_timedelta(self.config['timestep']).total_seconds()

    def _model_interface(self):
        """
        Interface for calling the different submodules. This method is called
        in every time step after the meteorological fields have been prepared.
        """
        modules.radiation.irradiance(self)

        if self.require_land_cover:
            self.land_cover.lai()

        if self.require_evapotranspiration or self.require_canopy:
            modules.canopy.above_canopy_meteorology(self)

        if self.require_canopy:
            self.canopy.meteorology()
            self.canopy.snow()

        self.snow.compaction()
        self.snow.accumulation()
        self.snow.albedo_aging()

        if self.config.snow_management.enabled:
            self.snow_management.produce()
            self.snow_management.groom()

        self.snow.update_properties()

        if self.require_soil:
            modules.soil.soil_properties(self)

        surface.surface_properties(self)

        if self.require_energy_balance:
            surface.energy_balance(self)
        elif self.require_temperature_index:
            self.snow.temperature_index()

        self.snow.heat_conduction()
        self.snow.melt()
        self.snow.sublimation()
        self.snow.runoff()
        self.snow.update_properties()
        self.snow.update_layers()

        if self.require_soil:
            modules.soil.soil_heat_flux(self)
            modules.soil.soil_temperature(self)

        if self.require_evapotranspiration:
            self.evapotranspiration.evapotranspiration()

    def _initialize_logger(self):
        """
        Initialize the logger for the model instance.
        """
        loguru.logger.remove()
        logger = copy.deepcopy(loguru.logger)
        log_format = ('<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | ' +
                      '<level>{message}</level>')
        logger.add(sys.stderr, format=log_format, filter='openamundsen', level=self.config.log_level)
        self.logger = logger

    def _initialize_grid(self):
        """
        Initialize the grid parameters (number of rows and columns, transformation
        parameters) for the OpenAmundsen instance by reading the DEM file associated to the
        model run.
        """
        self.logger.info('Initializing model grid')

        dem_file = util.raster_filename('dem', self.config)
        meta = fileio.read_raster_metadata(dem_file, crs=self.config.crs)
        self.logger.info(f'Grid has dimensions {meta["rows"]}x{meta["cols"]}')

        grid = util.ModelGrid(meta)
        grid.prepare_coordinates()
        self.grid = grid

    def _prepare_station_coordinates(self):
        """
        Transform the lon/lat coordinates of the meteorological stations to the
        coordinate system of the model grid. The transformed coordinates are
        stored in the `x` and `y` variables of the meteo dataset.
        Additionally, the row and column indices of the stations within the
        model grid are stored in the `row` and `col` variables, and two boolean
        variables `within_grid_extent` and `within_roi` indicate whether the
        stations lie within the model grid extent and the ROI, respectively.
        """
        ds = self.meteo

        x, y = util.transform_coords(ds.lon, ds.lat, constants.CRS_WGS84, self.config.crs)

        x_var = ds.lon.copy()
        x_var.values = x
        x_var.attrs = {
            'standard_name': 'projection_x_coordinate',
            'units': 'm',
        }

        y_var = ds.lat.copy()
        y_var.values = y
        y_var.attrs = {
            'standard_name': 'projection_y_coordinate',
            'units': 'm',
        }

        ds['x'] = x_var
        ds['y'] = y_var

        bool_var = ds.x.copy().astype(bool)
        bool_var[:] = False
        bool_var.attrs = {}
        int_var = bool_var.copy().astype(int)
        int_var[:] = -1

        rows, cols = rasterio.transform.rowcol(self.grid.transform, x, y)
        row_var = int_var.copy()
        col_var = int_var.copy()
        row_var.values[:] = rows
        col_var.values[:] = cols
        ds['col'] = col_var
        ds['row'] = row_var

        grid = self.grid
        ds['within_grid_extent'] = (
            (col_var >= 0)
            & (col_var < grid.cols)
            & (row_var >= 0)
            & (row_var < grid.rows)
        )

        within_roi_var = bool_var.copy()
        ds['within_roi'] = within_roi_var

        for station in ds.indexes['station']:
            dss = ds.sel(station=station)

            if dss.within_grid_extent:
                row = int(dss.row)
                col = int(dss.col)
                within_roi_var.loc[station] = self.grid.roi[row, col]

    def _initialize_state_variable_management(self):
        """
        Initialize the state variable management and add default state variables.
        """
        self.state = statevars.StateVariableManager(self.grid.rows, self.grid.cols)
        statevars.add_default_state_variables(self)

    def _create_state_variables(self):
        """
        Create the state variables (i.e., empty arrays with the shape of the
        model grid) for the model run.
        """
        self.state.initialize()

    def _initialize_state_variables(self):
        """
        Fill the state variables arrays with initial values.
        """
        self.snow.initialize()
        modules.soil.initialize(self)
        self.state.surface.temp[self.grid.roi] = self.state.soil.temp[0, self.grid.roi]

        if self.require_land_cover:
            self.land_cover.initialize()

        if self.require_canopy:
            self.canopy.initialize()

        if self.require_evapotranspiration:
            self.evapotranspiration.initialize()

        if self.config.snow_management.enabled:
            self.snow_management.initialize()

        if self.require_evapotranspiration:
            self.evapotranspiration.initialize()

    def _initialize_point_outputs(self):
        self.point_output = fileio.PointOutputManager(self)

    def _initialize_gridded_outputs(self):
        self.gridded_output = fileio.GriddedOutputManager(self)

    def _calculate_terrain_parameters(self):
        self.logger.info('Calculating terrain parameters')
        slope, aspect = terrain.slope_aspect(self.state.base.dem, self.grid.resolution)
        normal_vec = terrain.normal_vector(self.state.base.dem, self.grid.resolution)
        self.state.base.slope[:] = slope
        self.state.base.aspect[:] = aspect
        self.state.base.normal_vec[:] = normal_vec

    def _read_input_data(self):
        """
        Read the input raster files required for the model run including the
        DEM, ROI (if available), and other files depending on the activated
        submodules.
        """
        dem_file = util.raster_filename('dem', self.config)
        roi_file = util.raster_filename('roi', self.config)
        svf_file = util.raster_filename('svf', self.config)

        if dem_file.exists():
            self.logger.info(f'Reading DEM ({dem_file})')
            self.state.base.dem[:] = fileio.read_raster_file(dem_file, check_meta=self.grid)
        else:
            raise FileNotFoundError(f'DEM file not found: {dem_file}')

        if roi_file.exists():
            self.logger.info(f'Reading ROI ({roi_file})')
            self.grid.roi[:] = fileio.read_raster_file(roi_file, check_meta=self.grid)
        else:
            self.logger.debug('No ROI file available, setting ROI to entire grid area')
            self.grid.roi[:] = True

        if svf_file.exists():
            self.logger.info(f'Reading sky view factor ({svf_file})')
            self.state.base.svf[:] = fileio.read_raster_file(svf_file, check_meta=self.grid)
        else:
            self.logger.info('Calculating sky view factor')
            svf = terrain.sky_view_factor(
                self.state.base.dem,
                self.grid.resolution,
                num_sweeps=self.config.meteo.radiation.num_shadow_sweeps,
                logger=self.logger,
            )
            self.state.base.svf[:] = svf

            self.logger.debug(f'Writing sky view factor file ({svf_file})')
            fileio.write_raster_file(svf_file, svf, self.grid.transform)

        # Read snow redistribution factor files
        for precip_corr in self.config.meteo.precipitation_correction:
            if precip_corr['method'] == 'srf':
                if 'file' in precip_corr:
                    srf_file = precip_corr['file']
                else:
                    srf_file = util.raster_filename('srf', self.config)

                self.logger.info(f'Reading snow redistribution factor ({srf_file})')
                self.state.base.srf[:] = fileio.read_raster_file(srf_file, check_meta=self.grid)
                break

        # Read land cover file
        if self.require_land_cover:
            land_cover_file = util.raster_filename('lc', self.config)

            if land_cover_file.exists():
                self.logger.info(f'Reading land cover ({land_cover_file})')
                self.state.land_cover.land_cover[:] = fileio.read_raster_file(
                    land_cover_file,
                    check_meta=self.grid,
                )
            else:
                raise FileNotFoundError(f'Land cover file not found: {land_cover_file}')

        # Read soil texture file
        if self.require_soil_texture:
            soil_texture_file = util.raster_filename('soil', self.config)

            if soil_texture_file.exists():
                self.logger.info(f'Reading soil texture ({soil_texture_file})')
                self.state.evapotranspiration.soil_texture[:] = fileio.read_raster_file(
                    soil_texture_file,
                    check_meta=self.grid,
                )
            else:
                raise FileNotFoundError(f'Soil texture file not found: {soil_texture_file}')

        self.grid.prepare_roi_coordinates()

    def _read_meteo_data(self):
        """
        Read the meteorological data files required for the model run and store
        them in the `meteo` variable.
        """
        bounds = self.config.input_data.meteo.bounds
        if bounds == 'grid':
            x_min = self.grid.x_min
            y_min = self.grid.y_min
            x_max = self.grid.x_max
            y_max = self.grid.y_max
        elif bounds == 'global':
            x_min = -np.inf
            y_min = -np.inf
            x_max = np.inf
            y_max = np.inf
        elif isinstance(bounds, list):
            x_min, y_min, x_max, y_max = bounds

        if self.config.input_data.meteo.format == 'csv':
            if self.config.input_data.meteo.crs is None:
                meteo_crs = self.config.crs
            else:
                meteo_crs = self.config.input_data.meteo.crs
        elif self.config.input_data.meteo.format == 'netcdf':
            meteo_crs = None  # no CRS required for NetCDF input

        self.meteo = fileio.read_meteo_data(
            self.config.input_data.meteo.format,
            self.config.input_data.meteo.dir,
            self.config.start_date,
            self.config.end_date,
            meteo_crs=meteo_crs,
            grid_crs=self.config.crs,
            bounds=(x_min, y_min, x_max, y_max),
            exclude=self.config.input_data.meteo.exclude,
            include=self.config.input_data.meteo.include,
            freq=self.config['timestep'],
            aggregate=self.config.input_data.meteo.aggregate_when_downsampling,
            logger=self.logger,
        )

        self._prepare_station_coordinates()

        # reorder variables (only for aesthetic reasons)
        var_order = [
            'station_name',
            'lon',
            'lat',
            'alt',
            'x',
            'y',
            'col',
            'row',
            'within_grid_extent',
            'within_roi',
        ] + list(constants.METEO_VAR_METADATA.keys())
        self.meteo = self.meteo[var_order]

        meteo.correct_station_precipitation(self)

    def _process_meteo_data(self):
        """
        Calculate derived meteorological variables from the interpolated fields.
        """
        self.logger.debug('Calculating derived meteorological variables')

        m = self.state.meteo
        roi = self.grid.roi

        m.atmos_press[roi] = meteo.atmospheric_pressure(self.state.base.dem[roi])
        m.sat_vap_press[roi] = meteo.saturation_vapor_pressure(m.temp[roi])
        m.vap_press[roi] = meteo.vapor_pressure(m.temp[roi], m.rel_hum[roi])
        m.spec_hum[roi] = meteo.specific_humidity(m.atmos_press[roi], m.vap_press[roi])
        m.spec_heat_cap_moist_air[roi] = meteo.specific_heat_capacity_moist_air(m.spec_hum[roi])
        m.lat_heat_vap[roi] = meteo.latent_heat_of_vaporization(m.temp[roi])
        m.psych_const[roi] = meteo.psychrometric_constant(
            m.atmos_press[roi],
            m.spec_heat_cap_moist_air[roi],
            m.lat_heat_vap[roi],
        )
        m.wet_bulb_temp[roi] = meteo.wet_bulb_temperature(
            m.temp[roi],
            m.rel_hum[roi],
            m.vap_press[roi],
            m.psych_const[roi],
        )
        m.dew_point_temp[roi] = meteo.dew_point_temperature(m.temp[roi], m.rel_hum[roi])
        m.precipitable_water[roi] = meteo.precipitable_water(
            m.temp[roi],
            m.vap_press[roi],
        )
        m.dry_air_density[roi] = m.atmos_press[roi] / (constants.GAS_CONSTANT_DRY_AIR * m.temp[roi])

        # Calculate precipitation phase
        precip_phase_method = self.config.meteo.precipitation_phase.method
        if precip_phase_method == 'temp':
            pp_temp = m.temp
        elif precip_phase_method == 'wet_bulb_temp':
            pp_temp = m.wet_bulb_temp

        snowfall_frac = meteo.precipitation_phase(
            pp_temp[roi],
            threshold_temp=self.config.meteo.precipitation_phase.threshold_temp,
            temp_range=self.config.meteo.precipitation_phase.temp_range,
        )
        m.snowfall[roi] = snowfall_frac * m.precip[roi]
        m.rainfall[roi] = (1 - snowfall_frac) * m.precip[roi]

        # Redistribute snow
        if 'srf' in self.state.base:
            m.snowfall[roi] *= self.state.base.srf[roi]
            m.precip[roi] = m.snowfall[roi] + m.rainfall[roi]

        m.snowfall_rate[roi] = m.snowfall[roi] / self.timestep
        m.rainfall_rate[roi] = m.rainfall[roi] / self.timestep
        m.precip_rate[roi] = m.precip[roi] / self.timestep

    @property
    def is_first_timestep_of_model_run(self):
        return self.date_idx == 0

    @property
    def is_first_timestep_of_day(self):
        return (
            self.is_first_timestep_of_model_run
            or self.date.day != self.dates[self.date_idx - 1].day
        )


def Model(*args, **kwargs):
    import warnings
    warnings.warn(
        'Using oa.Model is deprecated, please use oa.OpenAmundsen instead',
        DeprecationWarning,
        stacklevel=2,
    )
    return OpenAmundsen(*args, **kwargs)



import numpy as np
from numpy.testing import assert_allclose, assert_equal
import openamundsen as oa
import pytest
import warnings


single_point_results_all = [
    pytest.lazy_fixture('single_point_results_multilayer'),
    pytest.lazy_fixture('single_point_results_cryolayers'),
]


@pytest.fixture(scope='module')
def point_results_cryolayers(base_config):
    config = base_config.copy()
    config.snow.model = 'cryolayers'
    config.output_data.timeseries.variables.append({'var': 'snow.cold_content'})
    config.output_data.timeseries.variables.append({'var': 'snow.layer_albedo'})
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    return model.point_output.data


@pytest.fixture(scope='function')
def single_point_results_multilayer(base_config_single_point_results):
    return base_config_single_point_results


@pytest.fixture(scope='function')
def single_point_results_cryolayers(point_results_cryolayers):
    return point_results_cryolayers.sel(point='proviantdepot')


def test_default_multilayer(base_config):
    assert base_config.snow.model == 'multilayer'


@pytest.mark.parametrize('ds', single_point_results_all)
def test_swe(ds):
    swe = ds.swe.values
    ice_content = ds.ice_content.values
    liquid_water_content = ds.liquid_water_content.values
    assert np.all(np.isfinite(swe))
    assert swe.min() == 0.
    assert_allclose(
        swe,
        (ice_content + liquid_water_content).sum(axis=1),
        atol=0.1,
    )
    assert np.all(ice_content >= 0.)
    assert np.all(liquid_water_content >= 0.)


def test_swe_max_mean(single_point_results_multilayer, single_point_results_cryolayers):
    swe = single_point_results_multilayer.swe.values
    assert_allclose(swe.max(), 123.95, atol=1)
    assert_allclose(swe.mean(), 67.38, atol=1)

    swe = single_point_results_cryolayers.swe.values
    assert_allclose(swe.max(), 109.6, atol=1)
    assert_allclose(swe.mean(), 56.29, atol=1)


@pytest.mark.parametrize('ds', single_point_results_all)
def test_density(ds):
    swe3d = (ds.ice_content + ds.liquid_water_content).values
    density = ds.snow_density.values
    assert np.all(np.isnan(density[swe3d == 0.]))
    assert np.all(np.isfinite(density[swe3d > 0.]))
    assert not np.any(density < 0.)
    assert not np.any(density > 1000.)


@pytest.mark.parametrize('ds', single_point_results_all)
def test_depth(ds):
    depth = ds.snow_depth.values
    swe = ds.swe.values
    assert np.all(depth[swe == 0.] == 0.)
    assert np.all(depth[swe > 0.] > 0.)
    assert not np.any(depth < 0.)

    swe3d = (ds.ice_content + ds.liquid_water_content).values
    thickness = ds.snow_thickness.values
    assert_allclose(depth, thickness.sum(axis=1), atol=1e-3)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            'ignore',
            'invalid value encountered in true_divide',
            RuntimeWarning,
        )
        density_calc = swe3d / thickness

    density = ds.snow_density.values
    assert_allclose(density_calc, density, atol=1.)


@pytest.mark.parametrize('ds', single_point_results_all)
def test_num_layers(ds):
    num_layers = ds.num_snow_layers.values
    thickness = ds.snow_thickness.values
    assert_equal(num_layers, (thickness > 0).sum(axis=1))


@pytest.mark.parametrize('ds', single_point_results_all)
def test_albedo(ds, base_config):
    albedo = ds.snow_albedo.values
    swe = ds.swe.values
    min_albedo = base_config.snow.albedo.min
    max_albedo = base_config.snow.albedo.max
    pos_snow = swe > 0.
    assert np.all(albedo[pos_snow] >= min_albedo)
    assert np.all(albedo[pos_snow] <= max_albedo)
    assert np.all(np.isnan(albedo[~pos_snow]))


@pytest.mark.parametrize('ds', single_point_results_all)
def test_melt(ds):
    assert np.all(ds.melt >= 0)


@pytest.mark.parametrize('ds', single_point_results_all)
def test_runoff(ds):
    assert np.all(ds.runoff >= 0)


def test_cryolayers_layer_albedo(single_point_results_cryolayers, base_config):
    ds = single_point_results_cryolayers
    layer_albedo = ds.layer_albedo.values
    swe3d = (ds.ice_content + ds.liquid_water_content).values
    min_albedo = base_config.snow.albedo.min
    max_albedo = base_config.snow.albedo.max
    pos_snow = swe3d > 0.
    assert np.all(layer_albedo[pos_snow] >= min_albedo)
    assert np.all(layer_albedo[pos_snow] <= max_albedo)
    assert np.all(np.isnan(layer_albedo[~pos_snow]))


def test_cryolayers_cold_content(single_point_results_cryolayers, base_config):
    ds = single_point_results_cryolayers
    cc = ds.cold_content.values
    ic = ds.ice_content.values
    lwc = ds.liquid_water_content.values
    melt = ds.melt.values
    sublimation = ds.sublimation.values
    swe3d = (ds.ice_content + ds.liquid_water_content).values
    cold_holding_capacity = base_config.snow.cryolayers.cold_holding_capacity
    pos_snow = swe3d > 0
    assert np.all(cc >= 0.)
    assert cc.max() > 0.
    assert np.all(cc[~pos_snow] == 0.)

    # Check if the CC is <= swe * cold_colding_capacity (add melt and sublimation because the
    # maximum CC is calculated before reducing SWE by melt and sublimation)
    assert np.all(
        (
            ((ic + lwc).sum(axis=1) + melt + sublimation) * cold_holding_capacity
            - cc.sum(axis=1)
        ) >= -0.1
    )



import numpy as np
from openamundsen import (
    constants,
    interpolation,
    meteo,
    modules,
)


def irradiance(model):
    model.sun_params = modules.radiation.sun_parameters(
        model.date,
        model.grid.center_lon,
        model.grid.center_lat,
        model.config.timezone,
    )

    clear_sky_shortwave_irradiance(model)
    shortwave_irradiance(model)
    longwave_irradiance(model)


def clear_sky_shortwave_irradiance(model):
    roi = model.grid.roi

    # Extend ROI with the station positions (when cloudiness is calculated via 'clear_sky_fraction',
    # clear sky irradiance should be available for all radiation stations, not only the inside-ROI
    # ones)
    station_rows = model.meteo.row[model.meteo.within_grid_extent].values
    station_cols = model.meteo.col[model.meteo.within_grid_extent].values
    roi_plus_stations = roi.copy()
    roi_plus_stations[station_rows, station_cols] = True

    mean_surface_albedo = model.state.surface.albedo[roi].mean()

    if model.sun_params['sun_over_horizon']:
        model.logger.debug('Calculating shadows')
        shadows = modules.radiation.shadows(
            model.state.base.dem,
            model.grid.resolution,
            model.sun_params['sun_vector'],
            num_sweeps=model.config.meteo.radiation.num_shadow_sweeps,
        )

        model.logger.debug('Calculating clear-sky shortwave irradiance')
        dir_irr, diff_irr = modules.radiation.clear_sky_shortwave_irradiance(
            model.sun_params['day_angle'],
            model.sun_params['sun_vector'],
            shadows,
            model.state.base.dem,
            model.state.base.svf,
            model.state.base.normal_vec,
            model.state.meteo.atmos_press,
            model.state.meteo.precipitable_water,
            mean_surface_albedo,
            roi=roi_plus_stations,
            ozone_layer_thickness=model.config.meteo.radiation.ozone_layer_thickness,
            atmospheric_visibility=model.config.meteo.radiation.atmospheric_visibility,
            single_scattering_albedo=model.config.meteo.radiation.single_scattering_albedo,
            clear_sky_albedo=model.config.meteo.radiation.clear_sky_albedo,
        )
    else:
        dir_irr = np.zeros((model.grid.rows, model.grid.cols))
        diff_irr = np.zeros((model.grid.rows, model.grid.cols))

    pot_irr = dir_irr + diff_irr
    model.state.meteo.sw_in_clearsky[roi_plus_stations] = pot_irr[roi_plus_stations]
    model.state.meteo.dir_in_clearsky[roi_plus_stations] = dir_irr[roi_plus_stations]
    model.state.meteo.diff_in_clearsky[roi_plus_stations] = diff_irr[roi_plus_stations]


def shortwave_irradiance(model):
    model.logger.debug('Calculating actual shortwave irradiance')
    cloud_config = model.config.meteo.interpolation.cloudiness
    roi = model.grid.roi
    m = model.state.meteo

    # Select stations within the grid extent and with shortwave radiation measurements
    # for the current time step
    ds_rad = (
        model.meteo
        .isel(station=model.meteo.within_grid_extent)
        .sel(time=model.date)
        .dropna('station', subset=['sw_in'])
    )
    num_rad_stations = len(ds_rad.station)

    # Select stations with temperature and humidity measurements for the current time step
    ds_temp_hum = (
        model.meteo
        .sel(time=model.date)
        .dropna('station', how='any', subset=['temp', 'rel_hum'])
    )
    num_temp_hum_stations = len(ds_temp_hum.station)

    method = 'constant'
    if model.sun_params['sun_over_horizon']:
        if cloud_config['day_method'] == 'clear_sky_fraction':
            if num_rad_stations > 0:
                method = 'clear_sky_fraction'
            else:
                if cloud_config['allow_fallback'] and num_temp_hum_stations > 0:
                    model.logger.debug('No radiation measurements available, falling back '
                                      'to humidity-based cloudiness calculation')
                    method = 'humidity'
                else:
                    model.logger.debug('No radiation measurements available, using cloudiness '
                                      'from previous time step')
                    method = 'constant'
        else:
            method = cloud_config['day_method']
    else:
        if cloud_config['night_method'] == 'humidity':
            if num_temp_hum_stations > 0:
                method = 'humidity'
            else:
                model.logger.debug('No temperature and humidity measurements available, using '
                                  'cloudiness from previous time step')
                method = 'constant'

    if method == 'constant':  # use cloudiness from previous time step
        cloud_factor_roi_prev = m.cloud_factor[roi]

        # When there is no cloudiness from the previous time step (i.e. when the model run starts
        # during nighttime) set cloudiness to a constant value
        cloud_factor_roi_prev[np.isnan(cloud_factor_roi_prev)] = 0.75

        m.cloud_factor[roi] = cloud_factor_roi_prev
    elif method == 'humidity':
        lr_t = model.config.meteo.interpolation.temperature.lapse_rate[model.date.month - 1]
        lr_td = model.config.meteo.interpolation.humidity.lapse_rate[model.date.month - 1]
        # TODO use here also the same settings for regression/fixed gradients as for the interpolation
        cloud_fracs = meteo.cloud_fraction_from_humidity(
            ds_temp_hum.temp,
            ds_temp_hum.rel_hum,
            ds_temp_hum.alt,
            lr_t,
            lr_td,
        )
        cloud_factors = meteo.cloud_factor_from_cloud_fraction(cloud_fracs)
        cloud_factor_interpol = interpolation.idw(
            ds_temp_hum.x,
            ds_temp_hum.y,
            cloud_factors,
            model.grid.roi_points[:, 0],
            model.grid.roi_points[:, 1],
        )
        m.cloud_factor[roi] = cloud_factor_interpol.clip(0, 1)
    elif method == 'clear_sky_fraction':
        cloud_factors = ds_rad.sw_in.values / m.sw_in_clearsky[ds_rad.row, ds_rad.col]
        cloud_factor_interpol = interpolation.idw(
            ds_rad.x,
            ds_rad.y,
            cloud_factors,
            model.grid.roi_points[:, 0],
            model.grid.roi_points[:, 1],
        )
        m.cloud_factor[roi] = cloud_factor_interpol.clip(0, 1)

    m.cloud_fraction[roi] = meteo.cloud_fraction_from_cloud_factor(m.cloud_factor[roi])
    m.sw_in[roi] = m.sw_in_clearsky[roi] * m.cloud_factor[roi]


def longwave_irradiance(model):
    model.logger.debug('Calculating longwave irradiance')
    roi = model.grid.roi
    m = model.state.meteo
    clear_sky_emissivity = meteo.clear_sky_emissivity(m.precipitable_water[roi])

    # TODO these should be parameters
    cloud_emissivity = 0.976  # emissivity of totally overcast skies (Greuell et al., 1997)
    rock_emission_factor = 0.01  # (K W-1 m2) temperature of emitting rocks during daytime is assumed to be higher than the air temperature by this factor multiplied by the incoming shortwave radiation (Greuell et al., 1997)

    # Incoming longwave radiation from the clear sky
    lw_in_clearsky = (
        clear_sky_emissivity
        * constants.STEFAN_BOLTZMANN
        * m.temp[roi]**4
        * model.state.base.svf[roi]
        * (1 - m.cloud_fraction[roi]**2)
    )

    # Incoming longwave radiation from clouds
    lw_in_clouds = (
        cloud_emissivity
        * constants.STEFAN_BOLTZMANN
        * m.temp[roi]**4
        * model.state.base.svf[roi]
        * m.cloud_fraction[roi]**2
    )

    # Incoming longwave radiation from surrounding slopes
    snowfree_count = (model.state.snow.swe[roi] == 0).sum()
    rock_fraction = snowfree_count / model.grid.roi.sum()
    lw_in_slopes = (
        constants.STEFAN_BOLTZMANN
        * (1 - model.state.base.svf[roi]) * (
            rock_fraction * (m.temp[roi] + rock_emission_factor * m.sw_in[roi])**4
            + (1 - rock_fraction) * model.state.surface.temp[roi]**4
        )
    )

    # Total incoming longwave radiation
    m.lw_in[roi] = lw_in_clearsky + lw_in_clouds + lw_in_slopes



import loguru
import numpy as np
import openamundsen as oa
from openamundsen import constants, errors, util
import pandas as pd
from pathlib import Path
import xarray as xr


def read_meteo_data(
    meteo_format,
    meteo_data_dir,
    start_date,
    end_date,
    meteo_crs=None,
    grid_crs=None,
    bounds=None,
    exclude=None,
    include=None,
    freq='H',
    aggregate=False,
    logger=None,
):
    """
    Read all available stations in NetCDF or CSV format for a given period.

    In case of NetCDF input, all available .nc files in `meteo_data_dir` are
    used.
    For CSV input a file named `stations.csv` must be provided in
    the specified directory containing the metadata (IDs, names, x/y
    coordinates, altitudes) of the stations. Station files must be in the same
    directory and must be named `<station_id>.csv`

    Parameters
    ----------
    meteo_format : str
        Data format (either 'netcdf' or 'csv').

    meteo_data_dir : path-like
        Location of the NetCDF files.

    start_date : datetime-like
        Start date.

    end_date : datetime-like
        End date.

    meteo_crs : str
        CRS of the station coordinates specified in the stations.csv file
        (required only when `meteo_format` is 'csv').

    grid_crs : str
        CRS of the model grid.

    bounds : list or None
        If specified, use only stations within the region specified as a list of
        (x_min, y_min, x_max, y_max) coordinates in the model grid CRS.

    exclude: list or None
        List of station IDs to exclude.

    include: list or None
        List of station IDs to include even if otherwise excluded via `bounds`
        or `exclude`.

    freq : str
        Pandas-compatible frequency string (e.g. '3H') to which the data should
        be resampled.

    aggregate : boolean, default False
        Aggregate data when downsampling to a lower frequency or take
        instantaneous values.

    logger : logger, default None
        Logger to use for status messages.

    Returns
    -------
    ds : Dataset
        Station data.
    """
    if logger is None:
        logger = loguru.logger

    meteo_data_dir = Path(meteo_data_dir)

    meta = _read_meteo_metadata(meteo_format, meteo_data_dir, meteo_crs, grid_crs)
    if meta.empty:
        raise errors.MeteoDataError('No stations found')

    meta = _apply_station_rules(meta, bounds, exclude, include)
    if meta.empty:
        raise errors.MeteoDataError('No stations available after applying station rules')

    datasets = []
    for station_id in meta.index:
        if meteo_format == 'netcdf':
            filename = meteo_data_dir / f'{station_id}.nc'
            logger.info(f'Reading meteo file: {filename}')
            ds = read_netcdf_meteo_file(filename, start_date=start_date, end_date=end_date)
        elif meteo_format == 'csv':
            filename = meteo_data_dir / f'{station_id}.csv'
            logger.info(f'Reading meteo file: {filename}')

            ds = read_csv_meteo_file(
                filename,
                station_id,
                meta.loc[station_id, 'name'],
                meta.loc[station_id, 'x'],
                meta.loc[station_id, 'y'],
                meta.loc[station_id, 'alt'],
                grid_crs,
            )
            ds = ds.sel(time=slice(start_date, end_date))

        if ds.time.to_index().inferred_freq is None:
            if ds.dims['time'] > 2:
                # ("> 2" because inferring the frequency requires at least 3 points, so for shorter
                # time series inferred_freq is always None)
                raise errors.MeteoDataError('Missing records or non-uniform timesteps in '
                                            f'{filename}')
        elif ds.time.to_index().inferred_freq != freq:
            ds = _resample_dataset(ds, freq, aggregate=aggregate)

        if ds.dims['time'] == 0:
            logger.warning('File contains no meteo data for the specified period')
        else:
            datasets.append(ds)

    if len(datasets) == 0:
        raise errors.MeteoDataError('No meteo data available for the specified period')

    ds_combined = _combine_meteo_datasts(datasets)
    dates_combined = ds_combined.time.to_index()
    if dates_combined[0] > start_date or dates_combined[-1] < end_date:
        raise errors.MeteoDataError('Insufficient meteo data available.\n'
                                    f'Requested period: {start_date}..{end_date}\n'
                                    f'Available period: {dates_combined[0]}..{dates_combined[-1]}')
    
    return ds_combined


def read_netcdf_meteo_file(filename, start_date=None, end_date=None):
    """
    Read a meteo data file in NetCDF format and
    - check if the time, lon, lat, and alt variables are included
    - rename the variables according to NETCDF_VAR_MAPPINGS
    - convert precipitation rates to sums if necessary
    - check if the units are as expected (METEO_VAR_METADATA)
    - remove all unsupported variables
    - set the station id and name as attributes.

    Parameters
    ----------
    filename : path-like
        CSV filename.

    Returns
    -------
    ds : Dataset
        Station data.
    """
    with xr.open_dataset(filename) as ds:
        ds = ds.sel(time=slice(start_date, end_date))
        ds.load()

    # rename variables
    ds_vars = list(ds.variables.keys())
    rename_vars = set(constants.NETCDF_VAR_MAPPINGS.keys()) & set(ds_vars)
    rename_dict = {v: constants.NETCDF_VAR_MAPPINGS[v] for v in rename_vars}
    ds = ds.rename_vars(rename_dict)

    for var in ('time', 'lon', 'lat', 'alt'):
        if var not in ds:
            raise oa.errors.MeteoDataError(f'File is missing "{var}" variable: {filename}')

    # Special case for precipitation - convert rates to sums first if required
    if 'precip' in ds and (
        'units' not in ds['precip'].attrs  # if units are not specified we assume they are kg m-2 s-1
        or ds['precip'].units == 'kg m-2 s-1'
    ):
        try:
            freq = pd.infer_freq(ds.indexes['time'])
            dt = util.offset_to_timedelta(freq).total_seconds()
        except ValueError:  # infer_freq() needs at least 3 dates
            dt = np.nan

        ds['precip'] *= dt
        ds['precip'].attrs['units'] = 'kg m-2'

    for var, meta in constants.METEO_VAR_METADATA.items():
        units = meta['units']

        if var in ds and 'units' in ds[var].attrs and ds[var].units != units:
            raise oa.errors.MeteoDataError(
                f'{var} has wrong units in {filename}: expected {units}, got {ds[var].units}'
            )

    # remove all unsupported variables
    ds_vars = list(ds.variables.keys())
    allowed_vars = ['time', 'lon', 'lat', 'alt'] + list(constants.NETCDF_VAR_MAPPINGS.values())
    drop_vars = list(set(ds_vars) - set(allowed_vars))
    ds = ds.drop_vars(drop_vars)

    station_id = _netcdf_station_id(filename)

    # set station name to station id (filename without extension) if not present
    if 'station_name' in ds.attrs:
        station_name = ds.attrs['station_name']
    else:
        station_name = station_id

    # remove all attributes except id and name
    ds.attrs = {
        'station_id': station_id,
        'station_name': station_name,
    }

    return ds


def read_csv_meteo_file(filename, station_id, station_name, x, y, alt, crs):
    """
    Read a meteorological data file in CSV format and return it as a Dataset.
    Unlike in read_netcdf_meteo_file, here it is assumed that precipitation is
    specified as a sum over the time step (i.e., kg m-2) instead of a rate (kg
    m-2 s-1).

    Parameters
    ----------
    filename : path-like
        CSV filename.

    station_id : str
        Station ID (must be unique).

    station_name : str
        Station name.

    x : float
        Station x coordinate.

    y : float
        Station y coordinate.

    alt : float
        Station altitude.

    crs : str
        CRS of the x/y coordinates.

    Returns
    -------
    ds : Dataset
        Station data.
    """
    param_mappings = {
        'temp': 'temp',
        'precip': 'precip',
        'rel_hum': 'rel_hum',
        'sw_in': 'sw_in',
        'wind_speed': 'wind_speed',
    }

    lon, lat = util.transform_coords(x, y, crs, constants.CRS_WGS84)

    df = pd.read_csv(filename, parse_dates=True, index_col=0)
    df = df.rename(columns=param_mappings)

    if 'precip' in df.columns:
        pass
        # TODO check if the time step is really 3600 s and if precip is really a sum

    datadict = {param: (['time'], df[param].astype(np.float32)) for param in df.columns}
    datadict.update(
        time=(['time'], df.index),
        lon=([], lon),
        lat=([], lat),
        alt=([], alt),
    )
    ds = xr.Dataset(datadict)

    ds.attrs = {
        'station_id': station_id,
        'station_name': station_name,
    }

    return ds


def _combine_meteo_datasts(datasets):
    """
    Combine a list of meteo datasets as read by read_netcdf_meteo_file.
    The datasets are merged by adding an additional "station" (= station id)
    dimension.
    """
    datasets_processed = []

    for ds in datasets:
        ds = ds.copy()

        for var, meta in constants.METEO_VAR_METADATA.items():
            # add missing variables to the dataset
            if var not in ds:
                ds[var] = xr.DataArray(
                    data=np.full(ds.time.shape, np.nan),
                    coords=ds.coords,
                    dims=('time',),
                )

            ds[var].attrs = {
                'standard_name': meta['standard_name'],
                'units': meta['units'],
            }

        ds.coords['station'] = ds.attrs['station_id']
        ds['station_name'] = ds.attrs['station_name']
        ds.attrs = {}

        datasets_processed.append(ds)

    return xr.combine_nested(datasets_processed, concat_dim='station')


def _resample_dataset(ds, freq, aggregate=False):
    """
    Resample a dataset to a given time frequency.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset.

    freq : str
        Pandas-compatible frequency string (e.g. '3H'). Must be an exact subset
        of the original frequency of the data.

    aggregate : boolean, default False
        Aggregate data when downsampling to a lower frequency or take
        instantaneous values.

    Returns
    -------
    ds_res : Dataset
        Resampled dataset.
    """
    if util.offset_to_timedelta(freq) >= pd.Timedelta('1d'):
        raise errors.MeteoDataError('Resampling to frequencies >= 1 day is not supported')

    # ds.resample() is extremely slow for some reason, so we resample using pandas
    df = ds.to_dataframe().drop(columns=['lon', 'lat', 'alt'])

    resample_kwargs = dict(label='right', closed='right', origin='start')

    if aggregate:
        # Calculate averages
        df_res = df.resample(freq, **resample_kwargs).mean()

        # We might end up with an extra bin after resampling; take only the dates which we would
        # have taken when using instantaneous values
        dates = df.asfreq(freq).index
        df_res = df_res.loc[dates]
    else:
        # Take the instantaneous values
        df_res = df.asfreq(freq)

    # Precipitation is summed up regardless of the aggregation setting
    if 'precip' in df:
        df_res['precip'] = df['precip'].resample(
            freq,
            **resample_kwargs,
        ).agg(pd.Series.sum, skipna=False)

    # Check if the desired frequency is a subset of the original frequency of the
    # data (e.g., resampling hourly to 3-hourly data is ok, but not hourly to
    # 1.5-hourly, or upsampling in general)
    if not df.index.intersection(df_res.index).equals(df_res.index):
        raise errors.MeteoDataError(f'Resampling from freq "{df.index.inferred_freq}" '
                                    f'to "{freq}" not supported')

    # Create a new dataset with the resampled time series
    ds_res = ds[['lon', 'lat', 'alt', 'time']]
    ds_res['time'] = df_res.index
    ds_res.attrs = ds.attrs

    for param in df_res.columns:
        ds_res[param] = df_res[param]

    return ds_res


def _netcdf_station_id(filename):
    """
    Return the station ID of a station in NetCDF format, i.e., the base name of
    the file without the .nc extension.
    """
    return Path(filename).stem


def _read_meteo_metadata(meteo_format, meteo_data_dir, meteo_crs, grid_crs):
    """
    Read the metadata of the available meteorological stations.

    Parameters
    ----------
    meteo_format : str
        Data format (either 'netcdf' or 'csv').

    meteo_data_dir : path-like
        Location of the NetCDF files.

    meteo_crs : str
        CRS of the station coordinates specified in the stations.csv file
        (required only when `meteo_format` is 'csv').

    grid_crs : str
        CRS of the model grid.

    Returns
    -------
    meta : DataFrame
        DataFrame containing the station IDs as index and the columns `name`,
        `x` (x coordinate in the grid CRS), `y` (y coordinate in the grid CRS),
        and `alt`.
    """
    if meteo_format == 'netcdf':
        nc_files = sorted(list(meteo_data_dir.glob('*.nc')))
        if len(nc_files) == 0:
            raise errors.MeteoDataError(f'No meteo data files found in {meteo_data_dir}')

        station_ids = []
        names = []
        lons = []
        lats = []
        alts = []

        for nc_file in nc_files:
            with xr.open_dataset(nc_file) as ds:
                station_ids.append(_netcdf_station_id(nc_file))
                names.append(None)  # name is not needed here, is set in read_netcdf_meteo_file
                lons.append(float(ds.lon.values))
                lats.append(float(ds.lat.values))
                alts.append(float(ds.alt.values))

        # Transform lon/lat coordinates to grid CRS
        xs_gridcrs, ys_gridcrs = util.transform_coords(
            lons,
            lats,
            constants.CRS_WGS84,
            grid_crs,
        )

        meta = pd.DataFrame(
            index=station_ids,
            data=dict(
                name=names,
                x=xs_gridcrs,
                y=ys_gridcrs,
                alt=alts,
            ),
        )
    elif meteo_format == 'csv':
        filename = f'{meteo_data_dir}/stations.csv'
        try:
            meta = pd.read_csv(filename, index_col=0)
        except FileNotFoundError:
            raise errors.MeteoDataError(f'Missing station metadata file ({filename})')

        required_cols = ['name', 'x', 'y', 'alt']
        if not all(pd.Index(required_cols).isin(meta.columns)):
            raise errors.MeteoDataError('Station metadata file does not contain all required '
                                        f'columns ({required_cols})')

        # Transform x/y coordinates from meteo CRS to grid CRS
        xs_gridcrs, ys_gridcrs = util.transform_coords(
            meta.x,
            meta.y,
            meteo_crs,
            grid_crs,
        )

        meta.x = xs_gridcrs
        meta.y = ys_gridcrs
    else:
        raise NotImplementedError

    return meta


def _apply_station_rules(meta, bounds, exclude, include):
    """
    Extend or reduce the list of stations to read in using a defined region or
    exclusion/inclusion lists.

    Parameters
    ----------
    meta : DataFrame
        Station metadata as returned by `_read_meteo_metadata`.

    bounds : list or None
        If specified, use only stations within the region specified as a list of
        (x_min, y_min, x_max, y_max) coordinates in the model grid CRS.

    exclude: list or None
        List of station IDs to exclude.

    include: list or None
        List of station IDs to include even if otherwise excluded via `bounds`
        or `exclude`.

    Returns
    -------
    meta : DataFrame
    """
    meta_all = meta.copy()

    if bounds is not None:
        # TODO check if using ">=" and "<" is correct here or if ">" and "<=" should be used for y
        # coordinates
        meta = meta[
            (meta.x >= bounds[0])
            & (meta.y >= bounds[1])
            & (meta.x < bounds[2])
            & (meta.y < bounds[3])
        ]

    if exclude is not None:
        meta = meta.drop(index=exclude, errors='ignore')

    if include is not None:
        for station_id in include:
            try:
                meta = meta.append(meta_all.loc[station_id])
            except KeyError:
                raise errors.MeteoDataError(f'Station to be included not found: {station_id}')

    meta = meta[~meta.index.duplicated(keep='first')]  # remove potential duplicate entries
    return meta



import openamundsen as oa
import copy
import pooch
import pytest
import textwrap

version = oa.__version__
if '+' in version:
    data_url = 'https://github.com/openamundsen/testdata/raw/{version}/'
else:
    data_url = 'https://github.com/openamundsen/testdata/raw/v{version}/'

data_fetcher = pooch.create(
    path=pooch.os_cache('openamundsen-tests'),
    base_url=data_url,
    version=version,
    version_dev='main',
    registry={  # hashes generated with generate_hashes.py
        'grids/rofental/dem_rofental_1000.asc': '368ee374bfbb927bae5408a38dbafbbeab4ff59dc94ae30956b617bf28c114ec',
        'grids/rofental/roi_rofental_1000.asc': '5c495833e8465931d7aedc982ef093332572034d9cf7da5fc8b98b72111e3d25',
        'grids/rofental/srf_rofental_1000.asc': '431f0ed2181abc2a79c6ffd79b804e5c983c9c47003bce57b853db2b41abf62d',
        'grids/rofental/svf_rofental_1000.asc': 'a4b7704fad6f7c63445a76c64207ff881b73023ec7d958d4581cd7fb42d06966',
        'meteo/rofental/csv/bellavista.csv': 'fab913929aa84e04a5c47c9c4ea89ed613b8ecc14cf54194a090600de4ce9d98',
        'meteo/rofental/csv/latschbloder.csv': '2d6a8a6297464d36ff9ccfc34ce4cd44a50f398c35de1a6e32e8922871503ff0',
        'meteo/rofental/csv/proviantdepot.csv': '7a068c921127c5e0f15ed3718c3a257c70146111ae8a1c5e97f0f8db1e60d2bc',
        'meteo/rofental/csv/stations.csv': '175b420d9659d3478d4755fc4b8d8c2569400b37d84ffb9906e1ab661afe5dc4',
        'meteo/rofental/netcdf/bellavista.nc': '96e5fc1af44f269015e52c2ec2a18f2e8d729401b5998d4fdbebb3b0014c8c58',
        'meteo/rofental/netcdf/latschbloder.nc': '0a885ebd2b70818a4e6c30498346827e61cd2e983a311eb9fc5b3257f5b01bd2',
        'meteo/rofental/netcdf/proviantdepot.nc': 'ecb70cdd0bbd1d87d77fcc0bebb3f28212b2aeae687380cf3a67c1b954f29c91',
    },
)
DATA_DIR = data_fetcher.abspath


def pytest_addoption(parser):
    parser.addoption('--run-slow', action='store_true', default=False, help='run slow tests')


def pytest_configure(config):
    pytest.DATA_DIR = DATA_DIR
    config.addinivalue_line('markers', 'slow: mark test as slow to run')


def pytest_collection_modifyitems(config, items):
    if config.getoption('--run-slow'):
        return

    skip_slow = pytest.mark.skip(reason='need --run-slow option to run')

    for item in items:
        if 'slow' in item.keywords:
            item.add_marker(skip_slow)


def pytest_sessionstart(session):
    fetch_data_files()


def fetch_data_files():
    for file in data_fetcher.registry_files:
        data_fetcher.fetch(file)


@pytest.fixture(scope='session')
def base_config():
    config_yaml = textwrap.dedent(f'''
        domain: rofental
        start_date: 2020-01-15
        end_date: 2020-04-30
        resolution: 1000
        timestep: 3H
        crs: "epsg:32632"
        timezone: 1

        input_data:
          grids:
            dir: {DATA_DIR}/grids/rofental
          meteo:
            dir: {DATA_DIR}/meteo/rofental/netcdf

        output_data:
          timeseries:
            format: memory

            variables:
              - var: snow.num_layers
                name: num_snow_layers
              - var: snow.albedo
                name: snow_albedo

          grids:
            format: memory
    ''')

    config = oa.Configuration.from_yaml(config_yaml)
    full_config = oa.parse_config(config)

    return full_config


@pytest.fixture(scope='session')
def base_config_point_results(base_config):
    model = oa.OpenAmundsen(base_config)
    model.initialize()
    model.run()
    return model.point_output.data


@pytest.fixture(scope='function')
def base_config_single_point_results(base_config_point_results):
    return base_config_point_results.sel(point='proviantdepot')


@pytest.fixture(scope='function')
def single_point_results_multilayer(base_config_single_point_results):
    return base_config_single_point_results



from munch import Munch
import numpy as np
from openamundsen import constants
import pandas as pd
import pandas.tseries.frequencies
from pathlib import Path, PosixPath, WindowsPath
import pyproj
import rasterio
import ruamel.yaml


class ConfigurationYAML(ruamel.yaml.YAML):
    def __init__(self):
        super().__init__(typ='rt')  # .indent() works only with the roundtrip dumper
        self.default_flow_style = False
        self.indent(mapping=2, sequence=4, offset=2)

        self.representer.add_representer(pd.Timestamp, self._repr_datetime)

        # Add representers for path objects (just using pathlib.Path does not
        # work, so have to add PosixPath and WindowsPath separately)
        self.representer.add_representer(PosixPath, self._repr_path)
        self.representer.add_representer(WindowsPath, self._repr_path)

    def _repr_datetime(self, representer, date):
        return representer.represent_str(str(date))

    def _repr_path(self, representer, path):
        return representer.represent_str(str(path))

    def dump(self, data, stream=None, **kw):
        inefficient = False

        if stream is None:
            inefficient = True
            stream = ruamel.yaml.compat.StringIO()

        super().dump(data, stream, **kw)

        if inefficient:
            return stream.getvalue()


yaml = ConfigurationYAML()


def create_empty_array(shape, dtype):
    """
    Create an empty array with a given shape and dtype initialized to "no
    data". The value of "no data" depends on the dtype and is e.g. NaN for
    float, 0 for int and False for float.

    Parameters
    ----------
    shape : int or sequence of ints
        Shape of the new array, e.g. (2, 3).

    dtype : type
        The desired data type for the array.

    Returns
    -------
    out : ndarray
    """
    dtype_init_vals = {
        float: np.nan,
        int: 0,
        bool: False,
    }

    return np.full(shape, dtype_init_vals[dtype], dtype=dtype)


def read_yaml_file(filename):
    """
    Read a YAML file.

    Parameters
    ----------
    filename : str

    Returns
    -------
    result : dict
    """
    with open(filename) as f:
        return load_yaml(f.read())


def load_yaml(s):
    return yaml.load(s)


def to_yaml(d):
    return yaml.dump(d)


def raster_filename(kind, config):
    """
    Return the filename of an input raster file for a model run.

    Parameters
    ----------
    kind : str
        Type of input file, e.g. 'dem' or 'roi'.

    config : dict
        Model run configuration.

    Returns
    -------
    file : pathlib.Path
    """
    dir = config['input_data']['grids']['dir']
    domain = config['domain']
    resolution = config['resolution']
    extension = 'asc'
    return Path(f'{dir}/{kind}_{domain}_{resolution}.{extension}')


def transform_coords(x, y, src_crs, dst_crs):
    """
    Transform coordinates from one coordinate reference system to another.

    Parameters
    ----------
    x : ndarray
        Source x coordinates or longitude.

    y : ndarray
        Source y coordinates or latitude.

    src_crs
        CRS of the original coordinates as a pyproj-compatible input
        (e.g. a string "epsg:xxxx").

    dst_crs
        Target CRS.

    Returns
    -------
    (x, y) tuple of ndarrays containing the transformed coordinates.
    """
    x = np.array(x)
    y = np.array(y)
    transformer = pyproj.Transformer.from_crs(src_crs, dst_crs, always_xy=True)
    return transformer.transform(x, y)


class ModelGrid(Munch):
    """
    Container for storing model grid related variables.
    """
    def prepare_coordinates(self):
        """
        Prepare a range of variables related to the grid coordinates:
        - xs, ys: 1d-arrays containing the x and y coordinates in the grid CRS.
        - X, Y, 2d-arrays containing the x and y coordinates for each grid point.
        - all_points: (N, 2)-array containing (x, y) coordinates of all grid points.
        - roi_points: (N, 2)-array containing (x, y) coordinates of all ROI points.
        - roi_idxs: (N, 2)-array containing (row, col) indexes of all ROI points.
        - roi_idxs_flat: 1d-array containing the flattened (1d) indexes of all ROI points
        """
        transform = self.transform
        if transform.a < 0 or transform.e > 0:
            raise NotImplementedError  # we only allow left-right and top-down oriented grids

        x_min = transform.xoff
        x_max = x_min + self.cols * self.resolution
        y_max = transform.yoff
        y_min = y_max - self.rows * self.resolution

        x_range, y_range = rasterio.transform.xy(
            self.transform,
            [0, self.rows - 1],
            [0, self.cols - 1],
        )
        xs = np.linspace(x_range[0], x_range[1], self.cols)
        ys = np.linspace(y_range[0], y_range[1], self.rows)
        X, Y = np.meshgrid(xs, ys)

        center_x = xs.mean()
        center_y = ys.mean()
        center_lon, center_lat = transform_coords(
            center_x,
            center_y,
            self.crs,
            constants.CRS_WGS84,
        )

        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max
        self.xs = xs
        self.ys = ys
        self.X = X
        self.Y = Y
        self.all_points = np.column_stack((X.flat, Y.flat))
        self.center_lon = center_lon
        self.center_lat = center_lat
        self.prepare_roi_coordinates()

    def prepare_roi_coordinates(self):
        """
        Update the roi_points and roi_idxs variables using the ROI field.
        """
        if 'roi' not in self:
            self.roi = np.ones((self.rows, self.cols), dtype=bool)

        roi_xs = self.X[self.roi]
        roi_ys = self.Y[self.roi]
        self.roi_points = np.column_stack((roi_xs, roi_ys))
        self.roi_idxs = np.array(np.where(self.roi)).T
        self.roi_idxs_flat = np.where(self.roi.flat)[0]


def offset_to_timedelta(offset):
    """
    Convert a pandas-compatible offset (e.g. '3H') to a Timedelta object.
    """
    return pd.to_timedelta(pandas.tseries.frequencies.to_offset(offset))



from .irradiance import irradiance
from .clearsky import clear_sky_shortwave_irradiance
from .shadows import shadows
from .sunparams import (
    day_angle,
    declination_angle,
    equation_of_time,
    hour_angle,
    sun_vector,
    sun_parameters,
)



import numpy as np
from numpy.testing import assert_allclose
import openamundsen as oa
from pathlib import Path


def test_evapotranspiration(base_config, tmp_path):
    config = base_config.copy()
    config.start_date = '2020-07-01'
    config.end_date = '2020-07-15'

    model = oa.OpenAmundsen(config)
    model.initialize()

    roi_xs = model.grid.X.flat[model.grid.roi_idxs_flat]
    roi_ys = model.grid.Y.flat[model.grid.roi_idxs_flat]

    for p in Path(config.input_data.grids.dir).glob('*.asc'):
        (tmp_path / p.name).symlink_to(p)

    shape = (model.grid.rows, model.grid.cols)
    soil = np.zeros(shape, dtype=int)
    lc = np.zeros(shape, dtype=int)

    # Test with all land cover classes and fixed soil texture
    soil[:] = 5
    lccs = model.config.land_cover.classes.keys()
    config.output_data.timeseries.points = []
    for lcc_num, lcc in enumerate(lccs):
        config.output_data.timeseries.points.append({
            'x': float(roi_xs[lcc_num]),
            'y': float(roi_ys[lcc_num]),
        })
        lc.flat[model.grid.roi_idxs_flat[lcc_num]] = lcc

    rio_meta = {'driver': 'AAIGrid'}
    oa.fileio.write_raster_file(
        oa.util.raster_filename('soil', config),
        soil.astype(np.int32),
        model.grid.transform,
        **rio_meta,
    )
    oa.fileio.write_raster_file(
        oa.util.raster_filename('lc', config),
        lc.astype(np.int32),
        model.grid.transform,
        **rio_meta,
    )

    config.evapotranspiration.enabled = True
    config.output_data.timeseries.add_default_points = False
    config.output_data.timeseries.variables = [
        {'var': 'evapotranspiration.evaporation'},
        {'var': 'evapotranspiration.transpiration'},
        {'var': 'evapotranspiration.evapotranspiration'},
    ]

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data

    for lcc_num, lcc in enumerate(lccs):
        lcc_params = model.config.land_cover.classes[lcc]
        crop_coeff_type = lcc_params.get('crop_coefficient_type', None)
        is_sealed = lcc_params.get('is_sealed', False)

        ds_lcc = ds.isel(point=lcc_num)
        assert np.all(ds_lcc.evapotranspiration >= 0)
        assert ds_lcc.evapotranspiration.max() > 0

        if crop_coeff_type == 'dual' or is_sealed:
            assert np.all(ds_lcc.evaporation >= 0)
            assert np.all(ds_lcc.transpiration >= 0)
            assert_allclose(
                ds_lcc.evaporation + ds_lcc.transpiration,
                ds_lcc.evapotranspiration,
                rtol=1e-3,
            )
        elif crop_coeff_type == 'single':
            assert np.all(np.isnan(ds_lcc.evaporation))
            assert np.all(np.isnan(ds_lcc.transpiration))

    # Test with all soil texture classes and fixed land cover
    lc[:] = 9
    soil[:] = 0
    stcs = range(1, 9 + 1)

    config.output_data.timeseries.points = []
    for stc_num, stc in enumerate(stcs):
        config.output_data.timeseries.points.append({
            'x': float(roi_xs[stc_num]),
            'y': float(roi_ys[stc_num]),
        })
        soil.flat[model.grid.roi_idxs_flat[stc_num]] = stc

    rio_meta = {'driver': 'AAIGrid'}
    oa.fileio.write_raster_file(
        oa.util.raster_filename('soil', config),
        soil.astype(np.int32),
        model.grid.transform,
        **rio_meta,
    )
    oa.fileio.write_raster_file(
        oa.util.raster_filename('lc', config),
        lc.astype(np.int32),
        model.grid.transform,
        **rio_meta,
    )

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data

    for stc_num, stc in enumerate(stcs):
        ds_stc = ds.isel(point=stc_num)
        assert np.all(ds_stc.evapotranspiration >= 0)
        assert ds_lcc.evapotranspiration.max() > 0
        assert np.all(ds_stc.evaporation >= 0)
        assert np.all(ds_stc.transpiration >= 0)
        assert_allclose(
            ds_stc.evaporation + ds_stc.transpiration,
            ds_stc.evapotranspiration,
            rtol=1e-3,
        )



import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
import openamundsen as oa
import openamundsen.errors as errors
import pandas as pd
from pathlib import Path
import pytest
import xarray as xr


def meteo_to_netcdf(ds, out_dir):
    ds = ds.copy(deep=True)
    td = ds.time[1] - ds.time[0]
    timestep = td.values / np.timedelta64(1, 's')

    ds = ds[[
        'station_name',
        'lon',
        'lat',
        'alt',
        'time',
        'temp',
        'precip',
        'rel_hum',
        'sw_in',
        'wind_speed',
    ]].rename_vars({
        'temp': 'tas',
        'precip': 'pr',
        'rel_hum': 'hurs',
        'sw_in': 'rsds',
        'wind_speed': 'wss',
    })

    ds['pr'] /= timestep
    ds['pr'].attrs = {
        'standard_name': 'precipitation_flux',
        'units': 'kg m-2 s-1',
    }

    for station_num in range(len(ds.station)):
        ds_cur = ds.isel(station=station_num).copy()
        station_id = str(ds_cur.station.values)
        station_name = str(ds_cur.station_name.values)
        ds_cur.attrs = {
            'Conventions': 'CF-1.6',
            'station_name': station_name,
        }
        ds_cur = ds_cur.drop_vars(['station', 'station_name'])
        ds_cur.to_netcdf(f'{out_dir}/{station_id}.nc')


def test_formats(base_config):
    config = base_config.copy()
    config.start_date = '2015-07-28'
    config.end_date = '2020-12-31'
    config.timestep = 'H'

    meta = pd.read_csv(f'{pytest.DATA_DIR}/meteo/rofental/csv/stations.csv', index_col=0)
    meta = meta.sort_index()

    meteo = {}
    for fmt in ('netcdf', 'csv'):
        config.input_data.meteo.format = fmt
        config.input_data.meteo.dir = f'{pytest.DATA_DIR}/meteo/rofental/{fmt}'
        model = oa.OpenAmundsen(config)
        model.initialize()
        ds = model.meteo.sortby('station')
        assert ds.time.to_index().equals(model.dates)
        assert_array_equal(meta['name'], ds.station_name)
        assert_array_equal(meta['alt'], ds.alt)
        meteo[fmt] = ds

    xr.testing.assert_allclose(meteo['netcdf'], meteo['csv'])


def test_no_files_found(base_config, tmp_path):
    config = base_config.copy()
    config.input_data.meteo.dir = str(tmp_path)

    for fmt in ('netcdf', 'csv'):
        config.input_data.meteo.format = fmt
        model = oa.OpenAmundsen(config)
        with pytest.raises(errors.MeteoDataError):
            model.initialize()


def test_missing_csv_metadata_columns(base_config, tmp_path):
    config = base_config.copy()
    config.input_data.meteo.format = 'csv'
    config.input_data.meteo.dir = str(tmp_path)

    p_orig = Path(f'{pytest.DATA_DIR}/meteo/rofental/csv')
    meta = pd.read_csv(p_orig / 'stations.csv', index_col=0)
    meta[['name', 'x', 'y']].loc[['bellavista']].to_csv(tmp_path / 'stations.csv')
    (tmp_path / 'bellavista.csv').symlink_to(p_orig / 'bellavista.csv')

    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_missing_records_inbetween(base_config, tmp_path):
    config = base_config.copy()
    config.timestep = 'H'
    config.input_data.meteo.format = 'csv'
    config.input_data.meteo.dir = str(tmp_path)

    p_orig = Path(f'{pytest.DATA_DIR}/meteo/rofental/csv')

    meta = pd.read_csv(p_orig / 'stations.csv', index_col=0)
    meta.loc[['bellavista']].to_csv(tmp_path / 'stations.csv')

    df = pd.read_csv(p_orig / 'bellavista.csv', index_col=0, parse_dates=True)
    df = df.loc[config.start_date:config.end_date]
    df = df.drop(df.index[37:38])
    df.to_csv(tmp_path / 'bellavista.csv')

    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_missing_records_start_end(base_config):
    config = base_config.copy()
    config.start_date = '1900-01-01'
    config.end_date = '2100-12-31'

    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_station_selection(base_config, tmp_path):
    config = base_config.copy()
    model = oa.OpenAmundsen(config)
    model.initialize()

    dummy_station = model.meteo.isel(station=0).copy()
    dummy_station['station'] = 'dummy'
    dummy_station['station_name'] = 'Dummy station'
    dummy_station['lon'] = 11.7
    dummy_station['lat'] = 47.21
    ds = xr.concat([model.meteo, dummy_station], dim='station')
    meteo_to_netcdf(ds, tmp_path)

    config.input_data.meteo.dir = str(tmp_path)
    config.input_data.meteo.bounds = 'grid'
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ['bellavista', 'latschbloder', 'proviantdepot'],
    )

    config.input_data.meteo.bounds = 'global'
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ['bellavista', 'dummy', 'latschbloder', 'proviantdepot'],
    )

    config.input_data.meteo.bounds = [
        636800,
        5182550,
        636850,
        5182580,
    ]
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ['bellavista'],
    )

    config.input_data.meteo.bounds = [
        636800,
        5182550,
        636801,
        5182551,
    ]
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    config.input_data.meteo.bounds = 'global'
    config.input_data.meteo.exclude = ['bellavista', 'latschbloder']
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ['dummy', 'proviantdepot'],
    )

    config.input_data.meteo.bounds = 'grid'
    config.input_data.meteo.exclude = ['bellavista', 'latschbloder', 'proviantdepot']
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    config.input_data.meteo.bounds = 'grid'
    config.input_data.meteo.exclude = []
    config.input_data.meteo.include = ['dummy']
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ['bellavista', 'dummy', 'latschbloder', 'proviantdepot'],
    )

    config.input_data.meteo.include = []
    config.input_data.meteo.exclude = ['dummy', 'dummy2']
    model = oa.OpenAmundsen(config)
    model.initialize()

    config.input_data.meteo.include = ['dummy', 'dummy2']
    config.input_data.meteo.exclude = []
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_missing_variables(base_config, tmp_path):
    ds = xr.load_dataset(f'{pytest.DATA_DIR}/meteo/rofental/netcdf/proviantdepot.nc')
    ds = ds.drop_vars(['hurs', 'rsds', 'wss'])
    ds.to_netcdf(tmp_path / 'proviantdepot.nc')

    config = base_config.copy()
    config.input_data.meteo.dir = str(tmp_path)
    model = oa.OpenAmundsen(config)
    model.initialize()

    assert 'rel_hum' in model.meteo
    assert 'sw_in' in model.meteo
    assert 'wind_speed' in model.meteo


def test_netcdf_precip_units(base_config, tmp_path):
    config = base_config.copy()
    config.input_data.meteo.dir = str(tmp_path)

    ds = xr.load_dataset(f'{pytest.DATA_DIR}/meteo/rofental/netcdf/proviantdepot.nc')
    ds.pr.attrs['units'] = 'kg m-2 s-1'
    ds.to_netcdf(tmp_path / 'proviantdepot.nc')
    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo1 = model.meteo.copy()

    ds['pr'] *= 3600
    ds.pr.attrs['units'] = 'kg m-2'
    ds.to_netcdf(tmp_path / 'proviantdepot.nc')
    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo2 = model.meteo.copy()

    xr.testing.assert_allclose(meteo1.precip, meteo2.precip)

    ds.pr.attrs['units'] = 'mm h-1'
    ds.to_netcdf(tmp_path / 'proviantdepot.nc')
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_crs(base_config, tmp_path):
    config = base_config.copy()

    model = oa.OpenAmundsen(config)
    model.initialize()
    ds1 = model.meteo[['lon', 'lat', 'x', 'y']].sortby('station')

    p_orig = Path(f'{pytest.DATA_DIR}/meteo/rofental/csv')
    meta = pd.read_csv(p_orig / 'stations.csv', index_col=0)
    x_new, y_new = oa.util.transform_coords(meta.x, meta.y, 'epsg:32632', 'epsg:3416')
    meta.x = x_new
    meta.y = y_new
    meta.to_csv(tmp_path / 'stations.csv')

    for station_id in meta.index:
        (tmp_path / f'{station_id}.csv').symlink_to(p_orig / f'{station_id}.csv')

    config.input_data.meteo.format = 'csv'
    config.input_data.meteo.dir = str(tmp_path)
    config.input_data.meteo.crs = 'epsg:3416'

    model = oa.OpenAmundsen(config)
    model.initialize()
    ds2 = model.meteo[['lon', 'lat', 'x', 'y']].sortby('station')

    xr.testing.assert_allclose(ds1, ds2)


def test_grid_cell_assignment(base_config, tmp_path):
    config = base_config.copy()
    config.start_date = '2020-01-01'
    config.end_date = '2020-01-01'

    model = oa.OpenAmundsen(config)
    model.initialize()
    grid_xs = model.grid.xs
    grid_ys = model.grid.ys

    config.input_data.meteo.format = 'csv'
    config.input_data.meteo.dir = str(tmp_path)

    station_coords_colrows = [  # (x, y, col, row)
        (grid_xs[0], grid_ys[0], 0, 0),
        (grid_xs[0] + 1, grid_ys[0] - 1, 0, 0),
        (grid_xs[1] + 1, grid_ys[0] - 1, 1, 0),
        (grid_xs[1] + 1, grid_ys[3] - 1, 1, 3),
        (grid_xs[-1] + 1, grid_ys[-1] - 1, len(grid_xs) - 1, len(grid_ys) - 1),
        (grid_xs[0], grid_ys[-1], 0, len(grid_ys) - 1),
    ]

    meta = pd.DataFrame(
        index=[f'dummy{i}' for i in range(len(station_coords_colrows))],
        data={
            'name': None,
            'x': [s[0] for s in station_coords_colrows],
            'y': [s[1] for s in station_coords_colrows],
            'alt': 0.,
        },
    )
    meta.to_csv(tmp_path / 'stations.csv')

    p_orig = Path(f'{pytest.DATA_DIR}/meteo/rofental/csv')
    for station_id in meta.index:
        (tmp_path / f'{station_id}.csv').symlink_to(p_orig / 'proviantdepot.csv')

    model = oa.OpenAmundsen(config)
    model.initialize()
    ds = model.meteo

    for i, s in enumerate(station_coords_colrows):
        ds_s = ds.sel(station=f'dummy{i}')
        expected_col = s[2]
        expected_row = s[3]
        assert int(ds_s.col) == expected_col
        assert int(ds_s.row) == expected_row


def test_resample(base_config, tmp_path):
    def meteo_to_df(model):
        return (
            model.meteo
            .sel(station='bellavista')
            .to_dataframe()
            [params]
        ).astype(float)

    params = ['temp', 'precip', 'rel_hum', 'sw_in', 'wind_speed']

    agg_funcs_inst = {p: lambda s: s.iloc[-1] for p in params}
    agg_funcs_inst['precip'] = pd.Series.sum
    agg_funcs_res = {p: pd.Series.mean for p in params}
    agg_funcs_res['precip'] = pd.Series.sum

    config = base_config.copy()
    config.start_date = '2015-11-20'
    config.end_date = '2015-11-30'
    config.timestep = 'H'
    model = oa.OpenAmundsen(config)
    model.initialize()
    df_h = meteo_to_df(model)

    config.timestep = '3H'
    model = oa.OpenAmundsen(config)
    model.initialize()
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc['2015-11-30 13:00':'2015-11-30 15:00'].agg(agg_funcs_inst, skipna=False),
        df_res.loc['2015-11-30 15:00'],
        check_exact=False,
        check_names=False,
    )
    assert np.isnan(df_res.loc['2015-11-26 15:00'].precip)
    assert_allclose(
        df_h.loc[:df_res.index[-1]].precip.sum(),
        df_res.precip.sum(),
    )

    config.input_data.meteo.aggregate_when_downsampling = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc['2015-11-30 13:00':'2015-11-30 15:00'].agg(agg_funcs_res, skipna=False),
        df_res.loc['2015-11-30 15:00'],
        check_exact=False,
        check_names=False,
    )
    assert np.isnan(df_res.loc['2015-11-26 15:00'].precip)

    config.timestep = 'D'
    config.input_data.meteo.aggregate_when_downsampling = True
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    config.start_date = '2015-11-01 01:00'
    config.end_date = '2015-11-30 22:00'
    config.timestep = '3H'
    config.input_data.meteo.aggregate_when_downsampling = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc['2015-11-30 14:00':'2015-11-30 16:00'].agg(agg_funcs_inst, skipna=False),
        df_res.loc['2015-11-30 16:00'],
        check_exact=False,
        check_names=False,
    )

    config.input_data.meteo.aggregate_when_downsampling = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc['2015-11-30 14:00':'2015-11-30 16:00'].agg(agg_funcs_res, skipna=False),
        df_res.loc['2015-11-30 16:00'],
        check_exact=False,
        check_names=False,
    )

    config.start_date = '2015-11-01 01:37'
    config.end_date = '2015-11-30 22:37'
    config.timestep = '3H'
    config.input_data.meteo.aggregate_when_downsampling = False
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    config.start_date = '2015-11-29'
    config.end_date = '2015-11-30'
    config.timestep = '10min'
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_non_hourly_input(base_config, tmp_path):
    config = base_config.copy()
    config.start_date = '2015-11-20'
    config.end_date = '2015-11-30'
    config.timestep = '3H'

    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo_to_netcdf(model.meteo, tmp_path)
    ds1 = model.meteo

    config.input_data.meteo.dir = str(tmp_path)
    model = oa.OpenAmundsen(config)
    model.initialize()
    ds2 = model.meteo

    xr.testing.assert_allclose(ds1, ds2)



from dataclasses import dataclass
import netCDF4
import numpy as np
from openamundsen import constants, errors, fileio, util
import pandas as pd
import pandas.tseries.frequencies
import pyproj
import xarray as xr


@dataclass
class OutputField:
    """
    Class for defining an output field, i.e., a state variable that should be
    written at specified dates.

    Parameters
    ----------
    var : str
        Name of the state variable (e.g. "meteo.temp").

    output_name : str
        Output name.

    agg : str, optional
        Aggregation function. Can be either None (if instantaneous values
        should be written), "sum" or "mean".

    write_dates : pd.DatetimeIndex
        Dates at which the field should be written.

    data : np.array, optional
        Current state of the aggregated values (only used if `agg` is not None).

    num_aggregations : int, default 0
        Current number of aggregations (required for calculating a running mean).
    """
    var: str
    output_name: str
    agg: str
    write_dates: pd.DatetimeIndex
    data: np.array = None
    num_aggregations: int = 0


def _field_key(field):
    return (tuple(field.write_dates), field.agg is None)


class GriddedOutputManager:
    """
    Class for managing and storing gridded output data which should be written at
    specified dates.

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.
    """
    def __init__(self, model):
        config = model.config.output_data.grids
        fields = []

        for field_cfg in config.variables:
            try:
                output_name = field_cfg['name']
            except KeyError:
                output_name = None

            try:
                freq = field_cfg['freq']
            except KeyError:
                freq = model.dates.freqstr

            try:
                agg = field_cfg['agg']
            except KeyError:
                agg = None

            if 'dates' in field_cfg:
                write_dates = pd.to_datetime(field_cfg['dates'])
            else:
                write_dates = _freq_write_dates(model.dates, freq, agg is not None)

            write_dates = write_dates[
                (write_dates >= model.dates[0])
                & (write_dates <= model.dates[-1])
            ]
            if len(write_dates) == 0:
                model.logger.debug(f'Discarding grid output variable {field_cfg["var"]}'
                                   ' (nothing to be written)')
                continue

            if output_name is None:
                output_name = field_cfg.var.split('.')[-1]

            if output_name in [f.output_name for f in fields]:
                raise errors.ConfigurationError(f'Duplicate grid output name: {output_name}')

            fields.append(OutputField(
                var=field_cfg['var'],
                output_name=output_name,
                agg=agg,
                write_dates=write_dates,
            ))

        self.model = model
        self.fields = fields
        self.format = config.format
        self.nc_file_created = False
        self.data = None

    def update(self):
        """
        Update the output fields for the current time step, i.e., update the
        aggregated fields (if aggregation functions are used) and write the
        variables to file at the specified dates.
        """
        # If there is nothing to be written, return right away
        if len(self.fields) == 0:
            return

        self.model.logger.debug('Updating field outputs')

        date = self.model.date
        roi = self.model.grid.roi

        if self.format == 'netcdf':
            nc_file = self.model.config.results_dir / 'output_grids.nc'

            if not self.nc_file_created:
                ds = self._create_dataset()
                ds.to_netcdf(nc_file)
                self.nc_file_created = True

            ds = netCDF4.Dataset(nc_file, 'r+')
        elif self.format == 'memory':
            if self.data is None:
                self.data = self._create_dataset()

        # Loop through all fields, update aggregations where necessary and write files at the
        # specified dates
        for field in self.fields:
            if field.agg is not None:
                if field.data is None:
                    meta = self.model.state.meta(field.var)

                    if meta.dim3 == 0:
                        arr = np.full((self.model.grid.rows, self.model.grid.cols), np.nan)
                        arr[roi] = 0
                    else:
                        arr = np.full((meta.dim3, self.model.grid.rows, self.model.grid.cols), np.nan)
                        arr[:, roi] = 0

                    field.data = arr

                data_cur = self.model.state[field.var]

                if field.agg == 'sum':
                    if field.data.ndim == 2:
                        field.data[roi] += data_cur[roi]
                    else:
                        field.data[:, roi] += data_cur[:, roi]
                elif field.agg == 'mean':
                    if field.data.ndim == 2:
                        field.data[roi] += (data_cur[roi] - field.data[roi]) / (field.num_aggregations + 1)
                    else:
                        field.data[:, roi] += (data_cur[:, roi] - field.data[:, roi]) / (field.num_aggregations + 1)

                field.num_aggregations += 1

            if date in field.write_dates:
                date_idx = np.flatnonzero(field.write_dates == date)[0]

                if field.agg is None:
                    data = self.model.state[field.var]
                else:
                    data = field.data

                if self.format == 'netcdf':
                    ds[field.output_name][date_idx, :, :] = data
                elif self.format in ('ascii', 'geotiff'):
                    if self.format == 'ascii':
                        ext = 'asc'
                        rio_meta = {'driver': 'AAIGrid'}
                        # (do not add CRS information when using AAIGrid output to avoid writing
                        # .prj files)
                    elif self.format == 'geotiff':
                        ext = 'tif'
                        rio_meta = {
                            'driver': 'GTiff',
                            'crs': self.model.grid.crs,
                        }

                    if field.agg is None:
                        date_str = f'{date:%Y-%m-%dT%H%M}'
                    else:
                        # Find the start date of the current output interval for the output file
                        # name
                        if date_idx == 0:
                            start_date = self.model.dates[0]
                        else:
                            start_date = field.write_dates[date_idx - 1] + pd.Timedelta(
                                seconds=self.model.timestep)

                        date_str = f'{start_date:%Y-%m-%dT%H%M}_{date:%Y-%m-%dT%H%M}'

                    if data.ndim == 2:
                        filename = self.model.config.results_dir / f'{field.output_name}_{date_str}.{ext}'
                        self.model.logger.debug(f'Writing field {field.var} to {filename}')
                        fileio.write_raster_file(
                            filename,
                            data,
                            self.model.grid.transform,
                            **rio_meta,
                        )
                    else:
                        # For 3-dimensional variables, write each layer as a separate file
                        for layer_num in range(data.shape[0]):
                            filename = (
                                self.model.config.results_dir
                                / f'{field.output_name}_{layer_num}_{date_str}.{ext}'
                            )
                            self.model.logger.debug(f'Writing field {field.var} (layer {layer_num})'
                                                    ' to {filename}')
                            fileio.write_raster_file(
                                filename,
                                data[layer_num, :, :],
                                self.model.grid.transform,
                                **rio_meta,
                            )
                elif self.format == 'memory':
                    self.data[field.output_name].values[date_idx, :, :] = data
                else:
                    raise NotImplementedError

                field.data = None
                field.num_aggregations = 0

        if self.format == 'netcdf':
            ds.close()

    def _create_dataset(self):
        """
        Create a CF-compliant Dataset covering the specified output variables
        and dates.

        Returns
        -------
        ds : xr.Dataset
        """
        # Define names of time variables - if there is only one time variable simply name it "time",
        # otherwise they are named "time1", "time2", ...
        time_var_names = {}
        num_time_vars = 0
        for field in self.fields:
            key = _field_key(field)
            if key not in time_var_names:
                num_time_vars += 1
                time_var_names[key] = f'time{num_time_vars}'

        if num_time_vars == 1:
            key = next(iter(time_var_names))
            time_var_names[key] = 'time'

        times = {}  # dict for storing times and boundaries (for aggregated variables) of the time variables
        field_time_vars = []  # contains for each field the name of the respective NetCDF time variable

        for field in self.fields:
            key = _field_key(field)
            time_var_name = time_var_names[key]
            time_vals = field.write_dates.values

            if field.agg is None:
                time_vals = field.write_dates
                time_bounds = None
            else:
                time_bounds = np.repeat(time_vals[:, np.newaxis], 2, axis=1).copy()
                time_bounds[1:, 0] = time_bounds[:-1, 1]
                time_bounds[0, 0] = self.model.dates[0]

            field_time_vars.append(time_var_name)

            if time_var_name not in times:
                times[time_var_name] = (time_vals, time_bounds)

        x_coords = self.model.grid.X[0, :]
        y_coords = self.model.grid.Y[:, 0]

        # Define coordinate variables
        coords = {}
        for time_var, (time_vals, time_bounds) in times.items():
            time_attrs = {}

            if time_bounds is not None:
                bound_var_name = f'{time_var}_bounds'
                time_attrs['bounds'] = bound_var_name
                coords[bound_var_name] = (
                    [time_var, 'nbnd'],
                    time_bounds,
                    {
                        'long_name': 'time interval endpoints',
                    }
                )

            coords[time_var] = (
                time_var,
                time_vals,
                time_attrs,
            )

        coords['x'] = (
            ['x'],
            x_coords,
            {
                'standard_name': 'projection_x_coordinate',
                'long_name': 'x coordinate of projection',
                'units': 'm',
            },
        )
        coords['y'] = (
            ['y'],
            y_coords,
            {
                'standard_name': 'projection_y_coordinate',
                'long_name': 'y coordinate of projection',
                'units': 'm',
            },
        )
        coords['crs'] = (
            [],
            np.array(0),
            pyproj.crs.CRS(self.model.grid.crs).to_cf(),
        )

        # Define data variables
        data = {}
        three_dim_coords = {}
        for field, field_time_var in zip(self.fields, field_time_vars):
            meta = self.model.state.meta(field.var)
            attrs = {}

            for attr in ('standard_name', 'long_name', 'units'):
                attr_val = getattr(meta, attr)
                if attr_val is not None:
                    attrs[attr] = attr_val

            attrs['grid_mapping'] = 'crs'

            # Assign output data type - float-like variables are written as float32, integer
            # variables as int32 or float32 (the latter if agg == 'mean')
            if (
                np.issubdtype(self.model.state.meta(field.var).dtype, np.integer)
                and field.agg != 'mean'
            ):
                dtype = np.int32
            else:
                dtype = np.float32

            if meta.dim3 == 0:  # 2-dimensional variable
                data[field.output_name] = (
                    [field_time_var, 'y', 'x'],
                    np.full((len(field.write_dates), len(y_coords), len(x_coords)), np.nan, dtype=dtype),
                    attrs,
                )
            else:  # 3-dimensional variable
                category = self.model.state.parse(field.var)[0]
                coord_name = f'{category}_layer'

                if category in three_dim_coords:
                    if three_dim_coords[coord_name] != meta.dim3:
                        # We assume that all 3-dimensional variables within a category have the
                        # same shape (e.g. "soil.temp" must have the same shape as "soil.therm_cond");
                        # varying numbers of layers within a category are not supported
                        raise Exception('Inconsistent length of third variable dimension')
                else:
                    three_dim_coords[coord_name] = meta.dim3

                data[field.output_name] = (
                    [field_time_var, coord_name, 'y', 'x'],
                    np.full(
                        (len(field.write_dates), meta.dim3, len(y_coords), len(x_coords)),
                        np.nan,
                        dtype=dtype,
                    ),
                    attrs,
                )

        # Add 3-dimensional coordinates
        for coord_name, coord_len in three_dim_coords.items():
            coords[coord_name] = ([coord_name], np.arange(coord_len))

        ds = xr.Dataset(data, coords=coords)
        ds.attrs['Conventions'] = 'CF-1.7'

        for time_var in times:
            ds[time_var].attrs['standard_name'] = 'time'

            # Set time units manually because otherwise the units of the time and the time bounds
            # variables might be different which is not recommended by CF standards
            ds[time_var].encoding['units'] = f'hours since {self.model.dates[0]:%Y-%m-%d %H:%M}'

            # Store time variables as doubles for CF compliance
            ds[time_var].encoding['dtype'] = np.float64
            if f'{time_var}_bounds' in ds:
                ds[f'{time_var}_bounds'].encoding['dtype'] = np.float64

        return ds


def _freq_write_dates(dates, out_freq, agg):
    """
    Calculate output dates for gridded outputs when a frequency string is set.

    For non-aggregated fields the write dates are assigned to the start of the
    respective intervals (e.g. if the model timestep is 1 hour and the write
    frequency is 'D', the write dates are 00:00 of each day). For aggregated
    fields, the write dates are assigned to the end of the intervals (e.g., in
    this case the write dates would be 23:00 of each day).

    Parameters
    ----------
    dates : pd.DatetimeIndex
        Simulation dates.

    out_freq : str
        Output frequency as a pandas offset string (e.g. '3H', 'M').

    agg : boolean
        Prepare write dates for aggregated outputs (if True) or for
        instantaneous values.

    Returns
    -------
    write_dates : pd.DatetimeIndex
    """
    model_freq = dates.freqstr
    model_freq_td = util.offset_to_timedelta(model_freq)
    out_offset = pandas.tseries.frequencies.to_offset(out_freq)

    if not out_offset.is_anchored():
        # For non-anchored offsets (e.g., '3H', 'D'), the output frequency must be a multiple of
        # (and not smaller than) the model timestep
        out_freq_td = util.offset_to_timedelta(out_freq)

        if out_freq_td < model_freq_td:
            raise ValueError('Output frequency must not be smaller than the model timestep')
        elif not (out_freq_td.total_seconds() / model_freq_td.total_seconds()).is_integer():
            raise ValueError('Output frequency must be a multiple of the model timestep')

    if agg:
        if out_offset.is_anchored():  # e.g. 'M', 'A'
            if model_freq_td.total_seconds() > constants.HOURS_PER_DAY * constants.SECONDS_PER_HOUR:
                raise NotImplementedError('Aggregation of gridded outputs with anchored offsets '
                                          'not supported for timesteps > 1d')

            period_end_dates = (
                pd.period_range(
                    start=dates[0],
                    end=dates[-1],
                    freq=out_freq,
                )
                .asfreq(model_freq, how='end')
                .to_timestamp()
            )

            d0 = dates[dates <= period_end_dates[0]][-1]
            write_dates = period_end_dates + (d0 - period_end_dates[0])

            if period_end_dates[0] - write_dates[0] > pd.Timedelta('1d'):
                write_dates = write_dates.delete(0)

            # Keep the last output interval only if it is fully covered (e.g., do not write half
            # months)
            if len(write_dates) > 0 and write_dates[-1] > dates[-1]:
                write_dates = write_dates.delete(-1)
        else:
            write_dates = pd.date_range(
                start=dates[0] + out_freq_td - model_freq_td,
                end=dates[-1],
                freq=out_freq,
            )
    else:
        write_dates = pd.date_range(
            start=dates[0],
            end=dates[-1],
            freq=out_freq,
        )

    return write_dates
